
module EMD 

	//////////////////PARAMETERS///////////////////////////

#( 
	parameter
	DATA_SIZE_BITS 		 = 16, 	//Number of bits representing every sample in the input signal.
	PRECISION_BITS_CS 	 = 5 , 	//Added precision bits for cubic spline calculation.
	PRECISION_BITS_MEMORY = 4 , 	//Precision bits used when storing (Maximum value : PRECISION_BITS_CS - 1).
	CHECKING_PRECISION 	 = 4 , 	//Used only for testing the performance when reducing the precision for the IMF_check (Can be ignored. Set to equal PRECISION_BITS_MEMORY)
	DEPTH 					 = 16, 	//Equivalent to the window size N 
	NUMBER_OF_IMFS 		 = 5		//Number of IMF the module should return. The module returns 0 if no more IMFs are available. 
)

	//////////////////INPUTS AND OUTPUTS////////////////////

(
	input 																	clk,
	input 																	rst_n,
	input 										[DATA_SIZE_BITS-1:0] data_in,
	input 																	data_in_available,
	output reg 									[DATA_SIZE_BITS-1:0] data_out,
	output reg  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] data_out_precision,
	output reg 																data_out_available,
	output reg 																EMD_finished
);

	//////////////////REGISTERS/////////////////////////////

	reg 			[DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] memory [0:DEPTH-1];
	reg 			[DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] data_in_sift;
	reg 											 [clogb2(DEPTH)-1:0] wr_ptr;
	reg 											 [clogb2(DEPTH)-1:0] rd_ptr;
	reg 											 [clogb2(DEPTH)-1:0] number_of_sifts;
	reg 																		read_out;
	reg 																		data_in_sift_available;
	reg 																		first_out;
	reg 																		last_out;
	reg 																		rst_n_sift;

	//////////////////WIRES////////////////////////////////

	wire	 	   [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] IMF_out;
	wire 		   [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] data_in_shifted;
	wire 																		IMF_out_available;

	//////////////////MODULES////////////////////////////////

	sift #(
			 DATA_SIZE_BITS,
			 PRECISION_BITS_CS,
			 PRECISION_BITS_MEMORY, 
			 CHECKING_PRECISION,
			 DEPTH
			)
			 
			 sift_module
			 
			(
			 clk, 
			 data_in_sift, 
			 data_in_sift_available, 
			 first_out, 
			 last_out,IMF_out, 
			 IMF_out_available, 
			 !(rst_n_sift&rst_n)
			);
			 
	//////////////////COMBINATORIAL////////////////////////

	assign data_in_shifted[DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] = data_in << PRECISION_BITS_MEMORY;


	//////////////////SEQUENTIAL///////////////////////////

	//-----------------------------------------------------
	//						MEMORY
	//-----------------------------------------------------
	always @(posedge clk) begin	
		if (data_in_available)
			memory[wr_ptr]= data_in_shifted;
		else if(IMF_out_available)
			memory[wr_ptr] = memory[wr_ptr] - IMF_out;
	end 

	
	//-----------------------------------------------------
	//						DATA_OUT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n| EMD_finished)begin 
			data_out <= 0;
			data_out_precision <=0;
		end
		else if(IMF_out_available)begin 
			data_out <= IMF_out[DATA_SIZE_BITS+PRECISION_BITS_MEMORY-1:PRECISION_BITS_MEMORY];
			data_out_precision <= IMF_out;
		end
	end
	
	
	//-----------------------------------------------------
	//						DATA_OUT_AVAILABLE
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n | EMD_finished)
			data_out_available <= 0;
		else
			data_out_available <= IMF_out_available;	
	end
	
	
	//-----------------------------------------------------
	//						DATA_IN_SIFT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n | EMD_finished)
			data_in_sift <= 0;
		else if(read_out)
			data_in_sift <= memory[rd_ptr];	
	end
	
	
	//-----------------------------------------------------
	//						DATA_IN_SIFT_AVAILABLE
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n | EMD_finished)
			data_in_sift_available <= 0;
		else
			data_in_sift_available <= read_out;
	end

	
	//-----------------------------------------------------
	//						FIRST_OUT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n | EMD_finished)
			first_out <= 0;
		else 
			first_out <= (rd_ptr == 0 & read_out);
	end
	

	//-----------------------------------------------------
	//						LAST_OUT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n | EMD_finished)
			last_out <= 0;
		else
			last_out <= rd_ptr == DEPTH-1;
	end
	
	
	//-----------------------------------------------------
	//						READ_POINTER
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n or posedge EMD_finished) begin
		if(!rst_n| EMD_finished)
			rd_ptr <= 0;
		else if(read_out)
			rd_ptr <= rd_ptr +1;	
	end

	
	//-----------------------------------------------------
	//						WRITE_POINTER
	//-----------------------------------------------------
	always @(posedge clk or negedge rst_n or posedge EMD_finished) begin
		if(!rst_n | EMD_finished)
			wr_ptr <= 0;
		else if(data_in_available | IMF_out_available)
			wr_ptr <= wr_ptr + 1;	
	end

	
	//-----------------------------------------------------
	//						READ_OUT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n| EMD_finished) 
			read_out <= 0;
		else if(wr_ptr == DEPTH-1)
			read_out <= 1;	
		else if(rd_ptr == DEPTH-1)
			read_out <= 0;
	end

	
	//-----------------------------------------------------
	//						EMD_FINISHED
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			EMD_finished <= 0;
		else if(number_of_sifts == NUMBER_OF_IMFS)
			EMD_finished <= 1;	
	end


	//-----------------------------------------------------
	//						NUMBER_OF_SIFTS
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n| EMD_finished)
			number_of_sifts <= 0;
		else if(IMF_out_available & (wr_ptr == DEPTH-1))
			number_of_sifts <= number_of_sifts+1;	
	end


	//-----------------------------------------------------
	//						RST_N_SIFT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n or posedge EMD_finished) begin
		if(!rst_n| EMD_finished)
			rst_n_sift <= 0;
		else if(IMF_out_available & (wr_ptr == DEPTH-1))
			rst_n_sift <= 0;
		else
			rst_n_sift <= 1;	
	end


	//-----------------------------------------------------
	//	Function for calculating number of bits needed to cover depth.
	//-----------------------------------------------------
		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction
endmodule


