
module calculate_spline 

	//////////////////PARAMETERS///////////////////////

#(
	parameter 
	DATA_SIZE_BITS 		 = 16, 
	PRECISION_BITS_CS 	 = 16, 
	PRECISION_BITS_MEMORY = 8, 
	DEPTH 					 = 512
)

	//////////////////INPUTS AND OUTPUTS//////////////

(
	input 																	 clk,
	input 																	 rst_n,
	input signed 	  [DATA_SIZE_BITS + 2*PRECISION_BITS_CS:0] a,
	input signed 	[DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] b,
	input signed 		 [DATA_SIZE_BITS + PRECISION_BITS_CS:0] c,
	input signed [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] d,
	input 										  [clogb2(DEPTH)-1:0] h,
	input 															 		 updated,
	input 																	 first,
	input 																	 last,
	output reg 	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] cubic_spline,
	output 																	 first_out,
	output 																	 last_out,
	output reg 																 data_out_available,
	output reg 																 calculate_spline_finished
);

	//////////////////REGISTERS///////////////////////

	reg 											  [clogb2(DEPTH)-1:0] counter;
	reg 																		 start;
	reg 																		 last_reg;

	//////////////////WIRES///////////////////////////
	
	wire 				  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:0] cubic_spline_aritmetic;
	wire 				  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:0] cubic_spline_aritmetic_rounded;

	//////////////////MODULES/////////////////////////
	
	round #(
			  DATA_SIZE_BITS + PRECISION_BITS_CS,
			  PRECISION_BITS_CS-PRECISION_BITS_MEMORY
			 )
			 
			  round_1
			  
			 (
			  cubic_spline_aritmetic,
			  cubic_spline_aritmetic_rounded
			 );

	//////////////////COMBINATORIAL///////////////////

	assign first_out = start & data_out_available;
	assign last_out  = last_reg & (counter==(h+1)); 
	assign cubic_spline_aritmetic = (((a*counter*counter*counter)>>> PRECISION_BITS_CS)+((b*counter*counter)>>>PRECISION_BITS_CS)+c*counter+(d<<(PRECISION_BITS_CS-PRECISION_BITS_MEMORY)));

	//////////////////SEQUENTIAL///////////////////////////

	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			cubic_spline <= 0;
		else
			cubic_spline <= cubic_spline_aritmetic_rounded[DATA_SIZE_BITS+PRECISION_BITS_MEMORY-1:0];	
	end

	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			last_reg <= 0;
		else if (last)
			last_reg <= 1;
		else if(last_out)
			last_reg <= 0;
	end


	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			counter <= 0;
		else if (calculate_spline_finished)
			counter <= 0;
		else if(counter < (h+last_reg))
			counter <= counter +1;
	end


	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			start <= 1;
		else if(last_out)
			start <= 1;
		else if(data_out_available)
			start <= 0;	

	end


	always @(posedge clk, negedge rst_n or posedge updated) begin
		if(!rst_n)
			calculate_spline_finished <= 1;
		else if(updated)
			calculate_spline_finished <= 0;
		else if((counter+(!last_reg)) >= h)
			calculate_spline_finished <= 1;	
	end


	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			data_out_available <= 0;
		else if(((counter > 0) & (counter< (h+last_reg))) | updated)
			data_out_available <= 1;
		else 
			data_out_available <= 0;
	end

	//Function for calculating number of bits needed to cover DEPTH.

		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction
endmodule

