


module calculate_spline_coefficients 

	//////////////////PARAMETERS///////////////////////////

#(
	parameter 
	DATA_SIZE_BITS = 16, 
	PRECISION_BITS_CS = 16, 
	PRECISION_BITS_MEMORY = 8, 
	DEPTH = 512
)

	//////////////////INPUTS AND OUTPUTS////////////////////

(
	input																		 clk,
	input 																	 rst_n,
	input signed [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] y,
	input 																	 y_available,
	input 																	 first_y,
	input			 								  [clogb2(DEPTH)-1:0] h,
	input																		 h_available,
	input signed 	  [(DATA_SIZE_BITS + PRECISION_BITS_CS):0] y_dd,
	input 																	 y_dd_available,
	input 																	 first,
	input 																	 last,
	output reg 		[(DATA_SIZE_BITS + 2*PRECISION_BITS_CS):0] a,
	output reg 	 [(DATA_SIZE_BITS + 2*PRECISION_BITS_CS)-1:0] b,
	output reg 		  [(DATA_SIZE_BITS + PRECISION_BITS_CS):0] c,
	output reg [(DATA_SIZE_BITS + PRECISION_BITS_MEMORY)-1:0] d,
	output reg 																 first_out,
	output reg 																 last_out,
	output reg 																 data_out_available
);

	//////////////////REGISTERS/////////////////////////////

	reg  [(DATA_SIZE_BITS + PRECISION_BITS_MEMORY)*DEPTH-1:0] y_reg;
	reg 									  [clogb2(DEPTH)*DEPTH-1:0] h_reg;
	reg 																		 initialing;
	reg 																		 initialing_delayed;
	reg signed 	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] y_m;
	reg signed 		  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:0] y_dd_delayed_one;
	reg signed 		  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:0] y_dd_delayed_two;

	reg signed 		[DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] y_dd_delayed_one_precision;
	reg signed 		[DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] y_dd_delayed_two_precision;
	reg 																		 last_delay_one;
	reg																		 last_delay_two;

	reg 																		 first_delay_one;
	reg 																		 first_delay_two;
	
	//////////////////WIRES////////////////////////////////

	wire 																		 full_fifo_h;
	wire 																		 full_fifo_y;
	wire 																		 empty_fifo_h;
	wire 																		 empty_fifo_y;
	wire 											[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_h;
	wire 											[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_y;
	wire 											  [clogb2(DEPTH)-1:0] fifo_out_h;
	wire signed  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] fifo_out_y;
	wire 																		 data_out_updated_fifo_h;
	wire 																		 data_out_updated_fifo_y;
	wire 																		 rd_en_fifo_y;
	wire 																		 rd_en_fifo_h;
	wire 																		 rst_n_no_fifo;

	wire signed 	  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:0] fifo_out_y_shifted;
	wire signed 	  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:0] y_m_shifted;
	wire signed    [DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] y_dd_delayed_two_div_2;
	wire signed 									 [clogb2(DEPTH):0] fifo_out_h_signed;
	
	//////////////////MODULES////////////////////////////////	

	FIFO #(
			 DATA_SIZE_BITS + PRECISION_BITS_MEMORY,
			 DEPTH/2
			)
			
			 fifo_y
			
			(
			 clk,
			 rd_en_fifo_y,
			 y_available,
			 full_fifo_y,
			 empty_fifo_y,
			 nr_elem_in_q_fifo_y,
			 y,
			 fifo_out_y,
			 data_out_updated_fifo_y,
			 rst_n
			);
			
	FIFO #(
			 clogb2(DEPTH),
			 DEPTH/2
			)
		
			 fifo_h	
			 
			(
			 clk,
			 rd_en_fifo_h,
			 h_available,
			 full_fifo_h, 
			 empty_fifo_h, 
			 nr_elem_in_q_fifo_h, 
			 h, 
			 fifo_out_h, 
			 data_out_updated_fifo_h, 
			 rst_n
			);	
	
	//////////////////COMBINATORIAL////////////////////////
	
	assign fifo_out_y_shifted [DATA_SIZE_BITS + PRECISION_BITS_CS-1:PRECISION_BITS_CS-PRECISION_BITS_MEMORY] = fifo_out_y;
	assign fifo_out_y_shifted 											  [PRECISION_BITS_CS-PRECISION_BITS_MEMORY-1:0] = 0;
	assign y_m_shifted 		  [DATA_SIZE_BITS + PRECISION_BITS_CS-1:PRECISION_BITS_CS-PRECISION_BITS_MEMORY] = y_m;
	assign y_m_shifted 													  [PRECISION_BITS_CS-PRECISION_BITS_MEMORY-1:0] = 0;
	assign y_dd_delayed_two_div_2 																									= y_dd_delayed_two_precision >>> 1;
	assign rd_en_fifo_y 																													= (initialing | y_dd_available | last_delay_one) & !empty_fifo_y;
	assign rd_en_fifo_h 																													= y_dd_available | last_delay_one;	
	assign fifo_out_h_signed 																				 [clogb2(DEPTH)-1:0] = fifo_out_h;
	assign fifo_out_h_signed 																					  [clogb2(DEPTH)] = 0;

	//////////////////SEQUENTIAL///////////////////////////
			
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			initialing <= 1;
		else if (last_out)
			initialing <= 1;
		else if(!empty_fifo_y & !data_out_available)
			initialing <= 0;
	end

	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			initialing_delayed <= 0;
		else
			initialing_delayed <= initialing;
	end


	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)		
			data_out_available <= 0;
		else
			data_out_available <= data_out_updated_fifo_y & !initialing_delayed &!last_out;	
	end

	
	//-----------------------------------------------------
	//						FIRST_OUT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n) begin
			first_delay_one <= 0;
			first_out <= 0;
		end
		else begin
			first_delay_one <= first;
			first_out <= first_delay_one; 
		end
	end
	
	
	//-----------------------------------------------------
	//						LAST_OUT
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n) begin
			last_delay_one <= 0;
			last_delay_two <= 0;
			last_out <= 0;
		end
		else begin
			last_delay_one <= last;
			last_delay_two <= last_delay_one;
			last_out <= last_delay_two; 
		end
	end


	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			y_m <= 0;
		else if(rd_en_fifo_y)
			y_m <= fifo_out_y;
		else if(last_out)
			y_m <= 0;
	end


	always @(posedge clk, negedge rst_n) begin
		if(!rst_n) begin
			y_dd_delayed_one <= 0;
			y_dd_delayed_two <= 0;
			y_dd_delayed_one_precision <= 0;
			y_dd_delayed_two_precision <= 0;		
			
		end
		else if(y_dd_available|last_delay_one) begin
			y_dd_delayed_one <= y_dd;
			y_dd_delayed_two <= y_dd_delayed_one;
			y_dd_delayed_one_precision <= y_dd << PRECISION_BITS_CS;
			y_dd_delayed_two_precision <= y_dd_delayed_one_precision;
		end
		else if(last_out)begin 
			y_dd_delayed_one <= 0;
			y_dd_delayed_two <= 0;
			y_dd_delayed_one_precision <= 0;
			y_dd_delayed_two_precision <= 0;
		end
		
	end
	
	
	//-----------------------------------------------------
	//						a, b, c, d
	//-----------------------------------------------------
	always @(posedge clk, negedge rst_n) begin
		if(!rst_n) begin
			a <= 0;
			b <= 0;
			c <= 0;
			d <= 0;
		end
		else if(last_delay_two) begin
			a <=  -y_dd_delayed_two_precision/(3*2*fifo_out_h_signed);
			b <= y_dd_delayed_two_div_2;
			c <= (fifo_out_y_shifted - y_m_shifted)/fifo_out_h_signed - y_dd_delayed_two*fifo_out_h_signed/3;
			d <= y_m;
		end
		else if(data_out_updated_fifo_y & !initialing_delayed) begin
			a <= (y_dd_delayed_one_precision- y_dd_delayed_two_precision)/(3*2*fifo_out_h_signed);
			b <= y_dd_delayed_two_div_2;
			c <= (fifo_out_y_shifted - y_m_shifted)/fifo_out_h_signed - y_dd_delayed_one*fifo_out_h_signed/(2*3) - y_dd_delayed_two*fifo_out_h_signed/3;
			d <= y_m;
		end
	end

	//Function for calculating number of bits needed to cover DEPTH.

		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction
		
	endmodule


