

module cubic_spline 

	//////////////////PARAMETERS///////////////////////////

#(
	parameter
	DATA_SIZE_BITS 		 = 16,
	PRECISION_BITS_CS 	 = 16, 
	PRECISION_BITS_MEMORY = 8, 
	DEPTH 					 = 512
)

	//////////////////INPUTS AND OUTPUTS//////////////


(
	input 															 clk,
	input 															 rst_n,
	input																 point_is_extreme,
	input 															 data_in_available,
	input 															 first,
	input 															 last,
	input  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] extremes,
	output [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] cubic_spline,
	output 															 first_out,
	output 															 last_out,
	output															 data_out_available
);

	//////////////////WIRES///////////////////////////

	wire 															  	 feeder_update;
	wire 															  	 feeder_first_out;
	wire 															  	 feeder_last_out;
	wire 									  						    feeder_distance_between_available;
	wire 	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] feeder_y_m;
	wire 	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] feeder_y;
	wire 	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] feeder_y_p;
	wire 									  [clogb2(DEPTH)-1:0] feeder_h;
	wire 									  [clogb2(DEPTH)-1:0] feeder_h_m;
	wire 									  [clogb2(DEPTH)-1:0] feeder_distance_between;
	
	wire 															  	 coef_data_out_available;
	wire 															 	 coef_first_out;
	wire 									   						 coef_last_out;
	wire 									    [clogb2(DEPTH):0] coef_A;
	wire 									  [clogb2(DEPTH)+2:0] coef_B;
	wire 							  		    [clogb2(DEPTH):0] coef_C;
	wire  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY +3:0] coef_D;

	wire 							 	 [PRECISION_BITS_CS-1:0] prime_C_prime;
	wire 			 [DATA_SIZE_BITS + PRECISION_BITS_CS:0] prime_D_prime;
	wire 															    prime_first_out;
	wire																 prime_last_out;
	wire 																 prime_data_out_available;

	wire 		  [(DATA_SIZE_BITS + PRECISION_BITS_CS):0] csd_y_dd;
	wire 																 csd_first_out;
	wire 																 csd_last_out;
	wire 																 csd_data_out_available;

	wire		  [DATA_SIZE_BITS + 2*PRECISION_BITS_CS:0] csc_a;
	wire		[DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] csc_b;
	wire			 [DATA_SIZE_BITS + PRECISION_BITS_CS:0] csc_c;
	wire	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] csc_d;
	wire 															 	 csc_first_out;
	wire 																 csc_last_out;
	wire 																 csc_data_out_available;


	wire		  [DATA_SIZE_BITS + 2*PRECISION_BITS_CS:0] FCS_a;
	wire		[DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] FCS_b;
	wire			 [DATA_SIZE_BITS + PRECISION_BITS_CS:0] FCS_c;
	wire	 [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] FCS_d;
	wire									  [clogb2(DEPTH)-1:0] FCS_h;
	wire 																 FCS_data_out_available;
	wire 																 FCS_first_out;
	wire																 FCS_last_out;
	
	wire 																 calculate_spline_finished;
	wire 																 reset_start;
	wire 																 reset_end;

	//////////////////MODULES/////////////////////////	
	
	feeder_cubic_spline_start 		  #(
												 DATA_SIZE_BITS, 
												 PRECISION_BITS_MEMORY,
												 DEPTH
												)
											
												 feeder_CS
												 
												(
												 clk,
												 reset_start,
												 point_is_extreme,
												 data_in_available,
												 first,
												 last,
												 extremes,
												 feeder_update,
												 feeder_first_out,
												 feeder_last_out,
												 feeder_y_m,
												 feeder_y,
												 feeder_y_p,
												 feeder_h,
												 feeder_h_m,
												 feeder_distance_between,
												 feeder_distance_between_available
												);

	calculate_coefficients 			  #(
												 DATA_SIZE_BITS, 
												 PRECISION_BITS_MEMORY, 
												 DEPTH
												) 
												
												 coef
												
												(
												 clk,
												 reset_start,
												 feeder_update,
												 feeder_first_out,
												 feeder_last_out,
												 feeder_y_m,
												 feeder_y,
												 feeder_y_p,
												 feeder_h_m,
												 feeder_h,
												 coef_data_out_available,
												 coef_first_out,
												 coef_last_out,
												 coef_A,
												 coef_B,
												 coef_C,
												 coef_D
												);


	calculate_primes					  #(
												 DATA_SIZE_BITS, 
												 PRECISION_BITS_CS, 
												 PRECISION_BITS_MEMORY, 
												 DEPTH
												 
												) 
												 prime
												 
												(
												 clk,
												 reset_start,
												 coef_A,
												 coef_B,
												 coef_C,
												 coef_D,
												 coef_data_out_available,
												 coef_first_out,
												 coef_last_out,
												 prime_C_prime,
												 prime_D_prime,
												 prime_first_out,
												 prime_last_out,
												 prime_data_out_available
												);




	calculate_second_derivative 	 #(
												DATA_SIZE_BITS, 
												PRECISION_BITS_CS, 
												DEPTH
											  ) 
												
												CSD
												
											  (
												clk,
												rst_n,
												prime_C_prime,
												prime_D_prime,
												prime_first_out,
												prime_last_out,
												prime_data_out_available,
												csd_y_dd,
												csd_first_out,
												csd_last_out,
												csd_data_out_available
											  );


	calculate_spline_coefficients #(
											  DATA_SIZE_BITS, 
											  PRECISION_BITS_CS, 
											  PRECISION_BITS_MEMORY, 
											  DEPTH
											 ) 
											 
											  CSC
											  
											 (
											  clk,
											  rst_n,
											  extremes,
											  point_is_extreme,
											  first,
											  feeder_distance_between,
											  feeder_distance_between_available,
											  csd_y_dd,
											  csd_data_out_available,
											  csd_first_out,
											  csd_last_out,
											  csc_a,
											  csc_b,
											  csc_c,
											  csc_d,
											  csc_first_out,
											  csc_last_out,
											  csc_data_out_available
											 );

	feeder_cubic_spline  			#(
											  DATA_SIZE_BITS, 
											  PRECISION_BITS_CS, 
											  PRECISION_BITS_MEMORY, 
											  DEPTH
											 ) 
											
											  FCS
											 
											 (
											  clk,
											  rst_n,
											  csc_a,
											  csc_b,
											  csc_c,
											  csc_d,
											  csc_data_out_available,
											  feeder_distance_between,
											  feeder_distance_between_available,
											  calculate_spline_finished,
											  FCS_a,
											  FCS_b,
											  FCS_c,
											  FCS_d,
											  FCS_h,
											  FCS_data_out_available,
											  FCS_first_out,
											  FCS_last_out
											 );

	calculate_spline  				#(
											  DATA_SIZE_BITS, 
											  PRECISION_BITS_CS, 
											  PRECISION_BITS_MEMORY, 
											  DEPTH
											 )
											
											  CS
											 
											 (
											  clk,
											  rst_n,
											  FCS_a,
											  FCS_b,
											  FCS_c,
											  FCS_d,
											  FCS_h,
											  FCS_data_out_available,
											  FCS_first_out,
											  FCS_last_out,
											  cubic_spline,
											  first_out,
											  last_out,
											  data_out_available,
											  calculate_spline_finished
											 );

	//////////////////COMBINATORIAL///////////////////
							
	assign reset_start = rst_n & !(first_out & data_out_available);
	assign reset_end   = rst_n & !(prime_last_out & prime_data_out_available);
							

	//Function for calculating number of bits needed to cover DEPTH.

		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction
endmodule
