
module feeder_cubic_spline 

	//////////////////PARAMETERS///////////////////////////

#( 
	parameter 
	DATA_SIZE_BITS 		 = 16, 
	PRECISION_BITS_CS 	 = 16, 
	PRECISION_BITS_MEMORY = 8, 
	DEPTH 					 = 512
)

	//////////////////INPUTS AND OUTPUTS////////////////////

(
	input																 clk,
	input																 rst_n,
	input  [DATA_SIZE_BITS + 2*PRECISION_BITS_CS:0] 	 a,
	input  [DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0]	 b,
	input  [DATA_SIZE_BITS + PRECISION_BITS_CS:0]		 c,
	input  [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] d,
	input 															 data_in_available,
	input 								  [clogb2(DEPTH)-1:0] h,
	input 															 h_available,
	input 															 calculate_spline_ready,
	output [DATA_SIZE_BITS + 2*PRECISION_BITS_CS:0] 	 a_out,
	output [DATA_SIZE_BITS + 2*PRECISION_BITS_CS-1:0] 	 b_out,
	output [DATA_SIZE_BITS + PRECISION_BITS_CS:0] 		 c_out,
	output [DATA_SIZE_BITS + PRECISION_BITS_MEMORY-1:0] d_out,
	output 								  [clogb2(DEPTH)-1:0] h_out,
	output 															 update,
	output 															 first_out,
	output 															 last_out
);

	//////////////////REGISTERS/////////////////////////////
	
	reg 																 start;
	
	//////////////////WIRES////////////////////////////////

	wire 																 rd_en;
	wire 																 full_fifo_a;
	wire																 full_fifo_b;
	wire																 full_fifo_c;
	wire																 full_fifo_d;
	wire																 full_fifo_h;
	wire																 empty_fifo_a;
	wire																 empty_fifo_b;
	wire																 empty_fifo_c;
	wire																 empty_fifo_d;
	wire																 empty_fifo_h;
	wire									[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_a;
	wire									[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_b;
	wire 									[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_c;
	wire 									[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_d;
	wire 									[clogb2(DEPTH/2)-1:0] nr_elem_in_q_fifo_h;
	wire 																 data_out_updated_fifo_a;
	wire																 data_out_updated_fifo_b;
	wire																 data_out_updated_fifo_c;
	wire																 data_out_updated_fifo_d;
	wire																 data_out_updated_fifo_h;

	//////////////////MODULES////////////////////////////////

	FIFO #(DATA_SIZE_BITS + 2*PRECISION_BITS_CS+1, DEPTH/2) fifo_a	(clk, rd_en, data_in_available, full_fifo_a, empty_fifo_a, nr_elem_in_q_fifo_a, a, a_out, data_out_updated_fifo_a, rst_n);
	FIFO #(DATA_SIZE_BITS + 2*PRECISION_BITS_CS,	  DEPTH/2) fifo_b	(clk, rd_en, data_in_available, full_fifo_b, empty_fifo_b, nr_elem_in_q_fifo_b, b, b_out, data_out_updated_fifo_b, rst_n);
	FIFO #(DATA_SIZE_BITS + PRECISION_BITS_CS+1,   DEPTH/2) fifo_c	(clk, rd_en, data_in_available, full_fifo_c, empty_fifo_c, nr_elem_in_q_fifo_c, c, c_out, data_out_updated_fifo_c, rst_n);
	FIFO #(DATA_SIZE_BITS + PRECISION_BITS_MEMORY, DEPTH/2) fifo_d	(clk, rd_en, data_in_available, full_fifo_d, empty_fifo_d, nr_elem_in_q_fifo_d, d, d_out, data_out_updated_fifo_d, rst_n);
	FIFO #(clogb2(DEPTH),								  DEPTH/2) fifo_h	(clk, rd_en, h_available		, full_fifo_h, empty_fifo_h, nr_elem_in_q_fifo_h, h, h_out, data_out_updated_fifo_h, rst_n);


	//////////////////COMBINATORIAL////////////////////////
	
	assign rd_en  	  = (nr_elem_in_q_fifo_a >=1) & calculate_spline_ready;
	assign update 	  = data_out_updated_fifo_a; 
	assign last_out  = !start & empty_fifo_a & update;
	assign first_out = start & update;

	//////////////////SEQUENTIAL///////////////////////////

	always @(posedge clk, negedge rst_n) begin
		if(!rst_n)
			start <= 1;
		else if (last_out)
			start <= 1;
		else if(data_out_updated_fifo_a)
			start <= 0;

	end

	//Function for calculating number of bits needed to cover DEPTH.

		function integer clogb2;
		input [31:0] depth;
			begin
				depth = depth-1;
				for(clogb2 = 0; depth>0; clogb2 = clogb2 +1)
					depth = depth >> 1;
			end
		endfunction
endmodule
